import typing as ta


DeployHome = ta.NewType('DeployHome', str)

DeployApp = ta.NewType('DeployApp', str)
DeployTag = ta.NewType('DeployTag', str)
DeployRev = ta.NewType('DeployRev', str)


class DeployAppTag(ta.NamedTuple):
    app: DeployApp
    tag: DeployTag



# ruff: noqa: UP006 UP007
import datetime
import os.path
import typing as ta

from .git import DeployGitManager
from .git import DeployGitRepo
from .git import DeployGitSpec
from .paths import DeployPath
from .paths import DeployPathOwner
from .types import DeployApp
from .types import DeployAppTag
from .types import DeployHome
from .types import DeployRev
from .types import DeployTag
from .venvs import DeployVenvManager


def make_deploy_tag(
        rev: DeployRev,
        now: ta.Optional[datetime.datetime] = None,
) -> DeployTag:
    if now is None:
        now = datetime.datetime.utcnow()  # noqa
    now_fmt = '%Y%m%dT%H%M%S'
    now_str = now.strftime(now_fmt)
    return DeployTag('-'.join([rev, now_str]))


class DeployAppManager(DeployPathOwner):
    def __init__(
            self,
            *,
            deploy_home: DeployHome,
            git: DeployGitManager,
            venvs: DeployVenvManager,
    ) -> None:
        super().__init__()

        self._deploy_home = deploy_home
        self._git = git
        self._venvs = venvs

        self._dir = os.path.join(deploy_home, 'apps')

    def get_deploy_paths(self) -> ta.AbstractSet[DeployPath]:
        return {
            DeployPath.parse('apps/@app/@tag'),
        }

    async def prepare_app(
            self,
            app: DeployApp,
            rev: DeployRev,
            repo: DeployGitRepo,
    ):
        app_tag = DeployAppTag(app, make_deploy_tag(rev))
        app_dir = os.path.join(self._dir, app, app_tag.tag)

        #

        await self._git.checkout(
            DeployGitSpec(
                repo=repo,
                rev=rev,
            ),
            app_dir,
        )

        #

        await self._venvs.setup_app_venv(app_tag)



# ruff: noqa: UP006 UP007
"""
TODO:
 - 'repos'?

git/github.com/wrmsr/omlish <- bootstrap repo
 - shallow clone off bootstrap into /apps

github.com/wrmsr/omlish@rev
"""
import dataclasses as dc
import functools
import os.path
import typing as ta

from omlish.lite.asyncio.subprocesses import asyncio_subprocess_check_call
from omlish.lite.cached import async_cached_nullary
from omlish.lite.check import check

from .paths import DeployPath
from .paths import DeployPathOwner
from .types import DeployHome
from .types import DeployRev


##


@dc.dataclass(frozen=True)
class DeployGitRepo:
    host: ta.Optional[str] = None
    username: ta.Optional[str] = None
    path: ta.Optional[str] = None

    def __post_init__(self) -> None:
        check.not_in('..', check.non_empty_str(self.host))
        check.not_in('.', check.non_empty_str(self.path))


@dc.dataclass(frozen=True)
class DeployGitSpec:
    repo: DeployGitRepo
    rev: DeployRev


##


class DeployGitManager(DeployPathOwner):
    def __init__(
            self,
            *,
            deploy_home: DeployHome,
    ) -> None:
        super().__init__()

        self._deploy_home = deploy_home
        self._dir = os.path.join(deploy_home, 'git')

        self._repo_dirs: ta.Dict[DeployGitRepo, DeployGitManager.RepoDir] = {}

    def get_deploy_paths(self) -> ta.AbstractSet[DeployPath]:
        return {
            DeployPath.parse('git'),
        }

    class RepoDir:
        def __init__(
                self,
                git: 'DeployGitManager',
                repo: DeployGitRepo,
        ) -> None:
            super().__init__()

            self._git = git
            self._repo = repo
            self._dir = os.path.join(
                self._git._dir,  # noqa
                check.non_empty_str(repo.host),
                check.non_empty_str(repo.path),
            )

        @property
        def repo(self) -> DeployGitRepo:
            return self._repo

        @property
        def url(self) -> str:
            if self._repo.username is not None:
                return f'{self._repo.username}@{self._repo.host}:{self._repo.path}'
            else:
                return f'https://{self._repo.host}/{self._repo.path}'

        async def _call(self, *cmd: str) -> None:
            await asyncio_subprocess_check_call(
                *cmd,
                cwd=self._dir,
            )

        @async_cached_nullary
        async def init(self) -> None:
            os.makedirs(self._dir, exist_ok=True)
            if os.path.exists(os.path.join(self._dir, '.git')):
                return

            await self._call('git', 'init')
            await self._call('git', 'remote', 'add', 'origin', self.url)

        async def fetch(self, rev: DeployRev) -> None:
            await self.init()
            await self._call('git', 'fetch', '--depth=1', 'origin', rev)

        async def checkout(self, rev: DeployRev, dst_dir: str) -> None:
            check.state(not os.path.exists(dst_dir))

            await self.fetch(rev)

            # FIXME: temp dir swap
            os.makedirs(dst_dir)

            dst_call = functools.partial(asyncio_subprocess_check_call, cwd=dst_dir)
            await dst_call('git', 'init')

            await dst_call('git', 'remote', 'add', 'local', self._dir)
            await dst_call('git', 'fetch', '--depth=1', 'local', rev)
            await dst_call('git', 'checkout', rev)

    def get_repo_dir(self, repo: DeployGitRepo) -> RepoDir:
        try:
            return self._repo_dirs[repo]
        except KeyError:
            repo_dir = self._repo_dirs[repo] = DeployGitManager.RepoDir(self, repo)
            return repo_dir

    async def checkout(self, spec: DeployGitSpec, dst_dir: str) -> None:
        await self.get_repo_dir(spec.repo).checkout(spec.rev, dst_dir)



# ruff: noqa: UP006 UP007
import dataclasses as dc

from omlish.lite.logs import log

from ..commands.base import Command
from ..commands.base import CommandExecutor


##


@dc.dataclass(frozen=True)
class DeployCommand(Command['DeployCommand.Output']):
    @dc.dataclass(frozen=True)
    class Output(Command.Output):
        pass


class DeployCommandExecutor(CommandExecutor[DeployCommand, DeployCommand.Output]):
    async def execute(self, cmd: DeployCommand) -> DeployCommand.Output:
        log.info('Deploying!')

        return DeployCommand.Output()



# ruff: noqa: UP006 UP007
"""
~deploy
  deploy.pid (flock)
  /app
    /<appspec> - shallow clone
  /conf
    /env
      <appspec>.env
    /nginx
      <appspec>.conf
    /supervisor
      <appspec>.conf
  /venv
    /<appspec>

?
  /logs
    /wrmsr--omlish--<spec>

spec = <name>--<rev>--<when>

==

for dn in [
    'app',
    'conf',
    'conf/env',
    'conf/nginx',
    'conf/supervisor',
    'venv',
]:

==

"""
import abc
import dataclasses as dc
import os.path
import typing as ta

from omlish.lite.check import check


DeployPathKind = ta.Literal['dir', 'file']  # ta.TypeAlias
DeployPathSpec = ta.Literal['app', 'tag']  # ta.TypeAlias


##


DEPLOY_PATH_SPEC_PLACEHOLDER = '@'
DEPLOY_PATH_SPEC_SEPARATORS = '-.'

DEPLOY_PATH_SPECS: ta.FrozenSet[str] = frozenset([
    'app',
    'tag',  # <rev>-<dt>
])


class DeployPathError(Exception):
    pass


@dc.dataclass(frozen=True)
class DeployPathPart(abc.ABC):  # noqa
    @property
    @abc.abstractmethod
    def kind(self) -> DeployPathKind:
        raise NotImplementedError

    @abc.abstractmethod
    def render(self, specs: ta.Optional[ta.Mapping[DeployPathSpec, str]] = None) -> str:
        raise NotImplementedError


#


class DirDeployPathPart(DeployPathPart, abc.ABC):
    @property
    def kind(self) -> DeployPathKind:
        return 'dir'

    @classmethod
    def parse(cls, s: str) -> 'DirDeployPathPart':
        if DEPLOY_PATH_SPEC_PLACEHOLDER in s:
            check.equal(s[0], DEPLOY_PATH_SPEC_PLACEHOLDER)
            return SpecDirDeployPathPart(s[1:])
        else:
            return ConstDirDeployPathPart(s)


class FileDeployPathPart(DeployPathPart, abc.ABC):
    @property
    def kind(self) -> DeployPathKind:
        return 'file'

    @classmethod
    def parse(cls, s: str) -> 'FileDeployPathPart':
        if DEPLOY_PATH_SPEC_PLACEHOLDER in s:
            check.equal(s[0], DEPLOY_PATH_SPEC_PLACEHOLDER)
            if not any(c in s for c in DEPLOY_PATH_SPEC_SEPARATORS):
                return SpecFileDeployPathPart(s[1:], '')
            else:
                p = min(f for c in DEPLOY_PATH_SPEC_SEPARATORS if (f := s.find(c)) > 0)
                return SpecFileDeployPathPart(s[1:p], s[p:])
        else:
            return ConstFileDeployPathPart(s)


#


@dc.dataclass(frozen=True)
class ConstDeployPathPart(DeployPathPart, abc.ABC):
    name: str

    def __post_init__(self) -> None:
        check.non_empty_str(self.name)
        check.not_in('/', self.name)
        check.not_in(DEPLOY_PATH_SPEC_PLACEHOLDER, self.name)

    def render(self, specs: ta.Optional[ta.Mapping[DeployPathSpec, str]] = None) -> str:
        return self.name


class ConstDirDeployPathPart(ConstDeployPathPart, DirDeployPathPart):
    pass


class ConstFileDeployPathPart(ConstDeployPathPart, FileDeployPathPart):
    pass


#


@dc.dataclass(frozen=True)
class SpecDeployPathPart(DeployPathPart, abc.ABC):
    spec: str  # DeployPathSpec

    def __post_init__(self) -> None:
        check.non_empty_str(self.spec)
        for c in [*DEPLOY_PATH_SPEC_SEPARATORS, DEPLOY_PATH_SPEC_PLACEHOLDER, '/']:
            check.not_in(c, self.spec)
        check.in_(self.spec, DEPLOY_PATH_SPECS)

    def _render_spec(self, specs: ta.Optional[ta.Mapping[DeployPathSpec, str]] = None) -> str:
        if specs is not None:
            return specs[self.spec]  # type: ignore
        else:
            return DEPLOY_PATH_SPEC_PLACEHOLDER + self.spec


@dc.dataclass(frozen=True)
class SpecDirDeployPathPart(SpecDeployPathPart, DirDeployPathPart):
    def render(self, specs: ta.Optional[ta.Mapping[DeployPathSpec, str]] = None) -> str:
        return self._render_spec(specs)


@dc.dataclass(frozen=True)
class SpecFileDeployPathPart(SpecDeployPathPart, FileDeployPathPart):
    suffix: str

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.suffix:
            for c in [DEPLOY_PATH_SPEC_PLACEHOLDER, '/']:
                check.not_in(c, self.suffix)

    def render(self, specs: ta.Optional[ta.Mapping[DeployPathSpec, str]] = None) -> str:
        return self._render_spec(specs) + self.suffix


##


@dc.dataclass(frozen=True)
class DeployPath:
    parts: ta.Sequence[DeployPathPart]

    def __post_init__(self) -> None:
        check.not_empty(self.parts)
        for p in self.parts[:-1]:
            check.equal(p.kind, 'dir')

        pd = {}
        for i, p in enumerate(self.parts):
            if isinstance(p, SpecDeployPathPart):
                if p.spec in pd:
                    raise DeployPathError('Duplicate specs in path', self)
                pd[p.spec] = i

        if 'tag' in pd:
            if 'app' not in pd or pd['app'] >= pd['tag']:
                raise DeployPathError('Tag spec in path without preceding app', self)

    @property
    def kind(self) -> ta.Literal['file', 'dir']:
        return self.parts[-1].kind

    def render(self, specs: ta.Optional[ta.Mapping[DeployPathSpec, str]] = None) -> str:
        return os.path.join(  # noqa
            *[p.render(specs) for p in self.parts],
            *([''] if self.kind == 'dir' else []),
        )

    @classmethod
    def parse(cls, s: str) -> 'DeployPath':
        tail_parse: ta.Callable[[str], DeployPathPart]
        if s.endswith('/'):
            tail_parse = DirDeployPathPart.parse
            s = s[:-1]
        else:
            tail_parse = FileDeployPathPart.parse
        ps = check.non_empty_str(s).split('/')
        return cls([
            *([DirDeployPathPart.parse(p) for p in ps[:-1]] if len(ps) > 1 else []),
            tail_parse(ps[-1]),
        ])


##


class DeployPathOwner(abc.ABC):
    @abc.abstractmethod
    def get_deploy_paths(self) -> ta.AbstractSet[DeployPath]:
        raise NotImplementedError



#!/usr/bin/env python3
# @omlish-amalg ../scripts/manage.py
# ruff: noqa: UP006 UP007
"""
manage.py -s 'docker run -i python:3.12'
manage.py -s 'ssh -i /foo/bar.pem foo@bar.baz' -q --python=python3.8
"""
import asyncio
import json
import sys
import typing as ta

from omlish.argparse.cli import ArgparseCli
from omlish.argparse.cli import argparse_arg
from omlish.argparse.cli import argparse_command
from omlish.lite.logs import log  # noqa
from omlish.lite.marshal import ObjMarshalerManager
from omlish.lite.marshal import ObjMarshalOptions
from omlish.lite.pycharm import PycharmRemoteDebug

from .bootstrap import MainBootstrap
from .bootstrap_ import main_bootstrap
from .commands.base import Command
from .config import MainConfig
from .deploy.config import DeployConfig
from .remote.config import RemoteConfig
from .targets.connection import ManageTargetConnector
from .targets.targets import ManageTarget


class MainCli(ArgparseCli):
    @argparse_command(
        argparse_arg('--_payload-file'),

        argparse_arg('--pycharm-debug-port', type=int),
        argparse_arg('--pycharm-debug-host'),
        argparse_arg('--pycharm-debug-version'),

        argparse_arg('--remote-timebomb-delay-s', type=float),

        argparse_arg('--debug', action='store_true'),

        argparse_arg('--deploy-home'),

        argparse_arg('target'),
        argparse_arg('command', nargs='+'),
    )
    async def run(self) -> None:
        bs = MainBootstrap(
            main_config=MainConfig(
                log_level='DEBUG' if self.args.debug else 'INFO',

                debug=bool(self.args.debug),
            ),

            deploy_config=DeployConfig(
                deploy_home=self.args.deploy_home,
            ),

            remote_config=RemoteConfig(
                payload_file=self.args._payload_file,  # noqa

                pycharm_remote_debug=PycharmRemoteDebug(
                    port=self.args.pycharm_debug_port,
                    **(dict(host=self.args.pycharm_debug_host) if self.args.pycharm_debug_host is not None else {}),
                    install_version=self.args.pycharm_debug_version,
                ) if self.args.pycharm_debug_port is not None else None,

                timebomb_delay_s=self.args.remote_timebomb_delay_s,
            ),
        )

        #

        injector = main_bootstrap(
            bs,
        )

        #

        msh = injector[ObjMarshalerManager]

        ts = self.args.target
        if not ts.startswith('{'):
            ts = json.dumps({ts: {}})
        tgt: ManageTarget = msh.unmarshal_obj(json.loads(ts), ManageTarget)

        cmds: ta.List[Command] = []
        cmd: Command
        for c in self.args.command:
            if not c.startswith('{'):
                c = json.dumps({c: {}})
            cmd = msh.unmarshal_obj(json.loads(c), Command)
            cmds.append(cmd)

        #

        async with injector[ManageTargetConnector].connect(tgt) as ce:
            async def run_command(cmd: Command) -> None:
                res = await ce.try_execute(
                    cmd,
                    log=log,
                    omit_exc_object=True,
                )

                print(msh.marshal_obj(res, opts=ObjMarshalOptions(raw_bytes=True)))

            await asyncio.gather(*[
                run_command(cmd)
                for cmd in cmds
            ])


def _main() -> None:
    sys.exit(asyncio.run(MainCli().async_cli_run()))


if __name__ == '__main__':
    _main()



# ruff: noqa: UP006 UP007
import os.path
import typing as ta

from omlish.lite.inject import InjectorBindingOrBindings
from omlish.lite.inject import InjectorBindings
from omlish.lite.inject import inj

from ..commands.inject import bind_command
from .apps import DeployAppManager
from .commands import DeployCommand
from .commands import DeployCommandExecutor
from .config import DeployConfig
from .git import DeployGitManager
from .interp import InterpCommand
from .interp import InterpCommandExecutor
from .types import DeployHome
from .venvs import DeployVenvManager


def bind_deploy(
        *,
        deploy_config: DeployConfig,
) -> InjectorBindings:
    lst: ta.List[InjectorBindingOrBindings] = [
        inj.bind(deploy_config),

        inj.bind(DeployAppManager, singleton=True),
        inj.bind(DeployGitManager, singleton=True),
        inj.bind(DeployVenvManager, singleton=True),

        bind_command(DeployCommand, DeployCommandExecutor),
        bind_command(InterpCommand, InterpCommandExecutor),
    ]

    if (dh := deploy_config.deploy_home) is not None:
        dh = os.path.abspath(os.path.expanduser(dh))
        lst.append(inj.bind(dh, key=DeployHome))

    return inj.as_bindings(*lst)



# ruff: noqa: UP006 UP007
"""
TODO:
 - interp
 - share more code with pyproject?
"""
import os.path
import typing as ta

from omlish.lite.asyncio.subprocesses import asyncio_subprocess_check_call

from .paths import DeployPath
from .paths import DeployPathOwner
from .types import DeployAppTag
from .types import DeployHome


class DeployVenvManager(DeployPathOwner):
    def __init__(
            self,
            *,
            deploy_home: DeployHome,
    ) -> None:
        super().__init__()

        self._deploy_home = deploy_home
        self._dir = os.path.join(deploy_home, 'venvs')

    def get_deploy_paths(self) -> ta.AbstractSet[DeployPath]:
        return {
            DeployPath.parse('venvs/@app/@tag/'),
        }

    async def setup_venv(
            self,
            app_dir: str,
            venv_dir: str,
            *,
            use_uv: bool = True,
    ) -> None:
        sys_exe = 'python3'

        await asyncio_subprocess_check_call(sys_exe, '-m', 'venv', venv_dir)

        #

        venv_exe = os.path.join(venv_dir, 'bin', 'python3')

        #

        reqs_txt = os.path.join(app_dir, 'requirements.txt')

        if os.path.isfile(reqs_txt):
            if use_uv:
                await asyncio_subprocess_check_call(venv_exe, '-m', 'pip', 'install', 'uv')
                pip_cmd = ['-m', 'uv', 'pip']
            else:
                pip_cmd = ['-m', 'pip']

            await asyncio_subprocess_check_call(venv_exe, *pip_cmd,'install', '-r', reqs_txt)

    async def setup_app_venv(self, app_tag: DeployAppTag) -> None:
        await self.setup_venv(
            os.path.join(self._deploy_home, 'apps', app_tag.app, app_tag.tag),
            os.path.join(self._deploy_home, 'venvs', app_tag.app, app_tag.tag),
        )
